
open "clauses_h";

(* Clauses C of the form:
 L1 \/ ... \/ Li \/ -Li+1 \/ ... \/ -Lj
 where each Lk is either of the form (1) P(x) or (2) P (f (x1,...,xn)), and either:
 (i) every Lk is of the form P(x) with the same x, or
 (ii) some Lk is of the form P (f (x1,...,xn)), and x1,...,xn contains all
 the free variables of C, and every atom of C of the form (2) is of the form
 Q (f (x1,...,xn)), with possibly different Q but the same f and xi's.

 Moreover, C is not a tautology.  So, in case (ii), every (2) atom must be
 on the same side of the clause, and must occur only once.  Every other atom
 can be grouped in a block, that is, a clause of the form (i).
 Moreover, we can normalize variable names:
 - in (i), by making it implicit;
 - in (ii), by numbering variables x1,...,xn from 1 to n, plus a canonical
 rewrite relation presenting equal variables (a flat union-find structure).
 e.g., f(x,y,x,z,z,z,y) would be f(1,2,3,4,5,6,7) plus rewrites 3->1, 5->4, 6->4, 7->2.
 This rewrite is coded as 1=>{3}, 2=>{7}, 4=>{5,6} (mapping i to set of j>i such
 that i and j are the same variable).

 Note no clause has any non-trivial factor, so binary (ordered) resolution
 is complete.
 We take an ordering that makes P (f (x1, ..., xn)) > Q (xi),
 and which is total on the Q (xi)'s, and which is total on the
 Q (f (x1,...,xn)) when Q varies (e.g., compare sizes of atoms first,
 then head predicate symbols).
 *)

fun maxvar_vars {} = 0
  | maxvar_vars {j} = j
  | maxvar_vars js =
    let val (js1, js2) = split js
    in
	max (maxvar_vars js1, maxvar_vars js2)
    end;

fun maxvar_eqvars {} = 0
  | maxvar_eqvars {i => js} =
    maxvar_vars js
  | maxvar_eqvars eqvars =
    let val (eqvars1, eqvars2) = split eqvars
    in
	max (maxvar_eqvars eqvars1, maxvar_eqvars eqvars2)
    end;

fun maxvar_blocks nil = 0
  | maxvar_blocks ((i, _)::rest) = max (i, maxvar_blocks rest);

val orsym = "; ";

fun print_clause (sigt, |[put, ...]|) =
    (fn C_BLOCK (pos, neg) =>
	let val delim = ref ""
	in
	    iterate
	      (put (!delim); delim := ", ";
	       put P; put "(x)")
	    | P in set neg
	    end;
	    if empty neg
		then ()
	    else put " => ";
		if empty pos
		    then put "0"
		else (delim := "";
		      iterate
			(put (!delim); delim := orsym;
			 put P; put "(x)")
		      | P in set pos
		      end)
	end
      | C_COMPLEX ((pos, neg), f, eqvars, blocks) => 
	let val n = if f inset sigt
			then ?sigt f
		    else max (maxvar_eqvars eqvars, maxvar_blocks blocks)
	    val varnums = iarray (n+1, 0)
	    (*$V-*)
	    val _ = iterate
		iterate
		  iupdate (varnums, j, i)
		| j in set js
		end
	    | i => js in map eqvars
	    end
	    (*$V+*)
	    memofun raw_varname j =
		    let val file as |[convert, ...]| = outstring "x"
		    in
			print file (pack j);
			convert ()
		    end
	    fun varname i =
		let val i' = isub (varnums, i)
		    val j = if i'=0 then i else i'
		in
		    raw_varname j
		end
	    val fname = let val |[put, convert, ...]| = outstring "("
			in
			    put f;
			    if n=0
				then put ")"
			    else let val delim = ref "("
				     val ir = ref 0
				 in
				     while (!ir<n) do
					 (put (!delim);
					  delim := ",";
					  inc ir;
					  put (varname (!ir)));
					 put "))"
				 end;
				 convert ()
			end
	    val delim = ref ""
	in
	    iterate
	      iterate
		(put (!delim); delim := ", ";
		 put P; put "("; put (raw_varname i); put ")")
	      | P in set neg
	      end
	    | (i, (_, neg)) in list blocks
	    end;
	    iterate
	      (put (!delim); delim := ", ";
	       put P; put fname)
	    | P in set neg
	    end;
	    if empty neg andalso
		all
		  empty neg
		| (_, (_, neg)) in list blocks
		end
		then ()
	    else put " => ";
		delim := "";
		iterate
		  (put (!delim); delim := orsym;
		   put P; put fname)
		| P in set pos
		end;
		iterate
		  iterate
		    (put (!delim); delim := orsym;
		     put P; put "("; put (raw_varname i); put ")")
		  | P in set pos
		  end
		| (i, (pos, _)) in list blocks
		end;
		if !delim=""
		    then put "0"
		else ()
	end
	);

fun print_id (file as |[put, ...]|) =
    fn (ID_NAME s) => (put "[Ax. "; put s; put "]")
     | (ID_NUM n) => (put "("; print file (pack n); put ")")
     | _ => put "(-)";

fun print_branch (sigt, file as |[put, ...]|) =
    let val prc = print_clause (sigt, file)
	val prid = print_id file
    in
	fn (BRANCH |[clauses, present, simple_queue, complex_queue, deductions, ...]|) =>
	   let fun prby (id1, id2, P) =
		   (put "\tby ";
		    prid id1;
		    put ", ";
		    prid id2;
		    put " on ";
		    put P)
	       fun prbyopt (J_DED by) = prby by
		 | prbyopt (J_AX id) = (put "\t"; prid id)
	       fun prbyid id =
		   if id inset deductions
		       then prby (?deductions id)
		   else ()
	   in
	       iterate
		 (if id inset present then () else put "(";
		      prid id; put "\t"; prc c; prbyid id;
		      if id inset present then () else put ")";
			  put "\n")
	       | id => c in map clauses
	       end;
	       if null simple_queue andalso null complex_queue
		   then ()
	       else (put "---------------------------------\n";
		     iterate
		       (prc c; prbyopt ded; put "\n")
		     | (c, ded) in list simple_queue
		     end;
		     iterate
		       let val delim = ref ""
		       in
			   iterate
			     (put (!delim); delim := " | ";
			      prc c)
			   | c in list cl
			   end;
			   prbyopt ded;
			   put "\n"
		       end
		     | (cl, ded) in list complex_queue
		     end)
	   end
    end;

fun print_tableau (sigt, file as |[put, ...]|) =
    let val prb = print_branch (sigt, file)
    in
	fn tab => let val n = len tab
		      val ir = ref 0
		      val tabr = ref tab
		  in
		      print file (pack n);
		      put " goal(s).\n";
		      while (!ir)<n do
			  (inc ir;
			   put "*** Goal ";
			   print file (pack (!ir));
			   put "\n";
			   prb (hd (!tabr));
			   tabr := tl (!tabr))
		  end
    end;


(* Given two relations between variables in ups1 and ups2
 (i.e., i ups1 j <=> j in ?ups1 i), compute composed relation
 ups1 U ups2 U (ups2 o ups1),
 where i (ups2 o ups1) j <=> exists k. i ups2 k and k ups1 j.
 *)
fun compose_eqvars (ups1, ups2) =
    ups1 ++
    {x => vars2 U union {vars1
			| y => vars1 in map vars2 <| ups1}
    | x => vars2 in map ups2};

(* Given two relations ups1 and ups2,
 compute closure of union ups defined by
 i ups j <=> exists k1,...,kn . i ups? k1 ... ups? kn ups? j
 where each ups? is either ups1 or ups2. *)
fun union_eqvars (ups1, ups2) =
    let fun iter ups =
	    let val ups' = compose_eqvars (ups, ups)
	    in
		if ups=ups'
		    then ups
		else iter ups'
	    end
    in
	iter (compose_eqvars (ups1, ups2))
    end;

exception TautologyEvt;
exception NoResolventEvt;

fun extract_blocks (_, _, nil) = nil
  | extract_blocks (blockr, vars, (ib as (i, block))::rest) =
    if i inset vars
	then let val (pos, neg) = !blockr
		 val (pos', neg') = block
		 val newpos = pos U pos'
		 val newneg = neg U neg'
	     in
		 if newpos intersects newneg
		     then raise TautologyEvt
		 else (blockr := (newpos, newneg);
		       extract_blocks (blockr, vars, rest))
	     end
    else ib::extract_blocks (blockr, vars, rest);

fun fuse_blocks1 (_, nil) = nil
  | fuse_blocks1 (eqvars, (ib as (i, block))::rest) =
    if i inset eqvars
	then let val blockr = ref block
		 val rest' = extract_blocks (blockr, ?eqvars i, rest)
	     in
		 (i, !blockr) :: fuse_blocks1 (eqvars, rest')
	     end
    else ib::fuse_blocks1 (eqvars, rest);

(* fuse_blocks (eqvars, blocks)
 given a unifier eqvars between variables, in the format of rewrites,
 and a list of blocks B1(x1) or ... or Bn(xn)
 returns (B1(x1) or ... or Bn(xn)) substituted by eqvars;
 this fuses blocks in general.
 *)
fun fuse_blocks ({}, blocks) = blocks
  | fuse_blocks (eqvars, blocks) = fuse_blocks1 (eqvars, blocks);

fun merge_blocks (nil, blocks2) = blocks2
  | merge_blocks (blocks1, nil) = blocks1
  | merge_blocks (blocks1 as (i1,b1)::rest1, blocks2 as (i2,b2)::rest2) =
    if i1=i2
	then let val pos = #1 b1 U #1 b2
		 val neg = #2 b1 U #2 b2
	     in
		 if pos intersects neg
		     then raise TautologyEvt
		 else (i1, (pos, neg))::merge_blocks (rest1, rest2)
	     end
    else if i1<i2
	then (i1,b1)::merge_blocks (rest1, blocks2)
    else (i2,b2)::merge_blocks (blocks1, rest2);

(* c_complex (pos, neg, f, eqvars, blocks) builds
 the complex clause
 +P1(f(x1..xk)) or ... or +Pn(f(x1..xk)) or -Q1(f(x1..xk)) or ... or -Qm(f(x1..xk)) or blocks
 where pos = {P1,...,Pn}, neg = {Q1,...,Qm}, and eqvars describes
 the variables in argument to f, x1,...,xk (some of which being equal);
 also, blocks is a list of blocks B_i(xi).
 If n=m=0, this is just blocks, the disjunction of the blocks B_i (x).
 Otherwise, this is a complex clause.
 *)
fun c_complex (pos, neg, f, eqvars, blocks) =
    if empty pos andalso empty neg
	then [C_BLOCK block | (_, block) in list blocks]
    else [C_COMPLEX ((pos, neg), f, eqvars, blocks)];

(* resolve (C1, C2, P) computes binary resolvent of C1 and C2 on some atom
 A with head P; C1 contains +A, C2 contains -A.
 *)
fun resolve (C_BLOCK (pos1, neg1), C_BLOCK (pos2, neg2), P) =
    let val pos = (pos1 \ {P}) U pos2
	val neg = neg1 U (neg2 \ {P})
    in
	if pos intersects neg
	    then raise TautologyEvt
	else [C_BLOCK (pos, neg)]
    end
  | resolve (C_BLOCK (pos1, neg1), C_COMPLEX ((pos2, neg2), f, eqvars, blocks), P) =
    let val pos = (pos1 \ {P}) U pos2
	val neg = neg1 U (neg2 \ {P})
    in
	if pos intersects neg
	    then raise TautologyEvt
	else c_complex (pos, neg, f, eqvars, blocks)
    end
  | resolve (C_COMPLEX ((pos1, neg1), f, eqvars, blocks), C_BLOCK (pos2, neg2), P) =
    let val pos = (pos1 \ {P}) U pos2
	val neg = neg1 U (neg2 \ {P})
    in
	if pos intersects neg
	    then raise TautologyEvt
	else c_complex (pos, neg, f, eqvars, blocks)
    end
  | resolve (C_COMPLEX ((pos1, neg1), f1, eqvars1, blocks1),
	     C_COMPLEX ((pos2, neg2), f2, eqvars2, blocks2), P) =
    if f1=f2
	then let val pos = (pos1 \ {P}) U pos2
		 val neg = neg1 U (neg2 \ {P})
	     in
		 if pos intersects neg
		     then raise TautologyEvt
		 else let val neweqs = union_eqvars (eqvars1, eqvars2)
			  val blocks'1 = if eqvars1=neweqs
					     then blocks1
					 else fuse_blocks (neweqs, blocks1)
			  val blocks'2 = if eqvars2=neweqs
					     then blocks2
					 else fuse_blocks (neweqs, blocks2)
		      in
			  c_complex (pos, neg, f1, neweqs,
				     merge_blocks (blocks'1, blocks'2))
		      end
	     end
    else raise NoResolventEvt;


exception MaxPreds;

fun maximal_literal less =
    let memofun maxpreds {} = raise MaxPreds
	      | maxpreds {P} = P
	      | maxpreds Ps = let val (P1s, P2s) = split Ps
				  val P1 = maxpreds P1s
				  val P2 = maxpreds P2s
			      in
				  if less (P1, P2)
				      then P2
				  else P1
			      end
	memofun maxblock ({}, neg) = (maxpreds neg, false)
	      | maxblock (pos, {}) = (maxpreds pos, true)
	      | maxblock (pos, neg) =
		let val P1 = maxpreds pos
		    val P2 = maxpreds neg
		in
		    if less (P1, P2)
			then (P2, false)
		    else (P1, true)
		end
	fun maxclause (C_BLOCK b) = (maxblock b, NONE)
	  | maxclause (C_COMPLEX (b, f, ...)) = (maxblock b, SOME f)
    in
	maxclause
    end;

(* Note: the rule we use when we generate a disjunction of blocks
 is a tableau beta-rule.
 Another possibility [Joyner76] would be to do a special form
 of factoring, whereby C is replaced by C.sigma whenever C.sigma
 is a factor of C that is included in C.
 E.g., P(x)\/P(y) can be replaced by P(x)  (sigma = [x:=y]).
 This way, on n predicates, we have at most 2^n blocks, then
 at most 2^{2^n} disjunctions of blocks.  This is of the same
 worst-case complexity than the tableau rule, however the tableau
 rule seems more natural.  Hopefully it is also more efficient
 in practical cases.
 *)

(*
datatype 'a mutlist = Mnil | Mcons of 'a * 'a mutlist ref;
datatype 'a queue = QUEUE of 'a mutlist ref (* head of the queue *)
    * 'a mutlist ref ref (* tail of the queue. *);

exception QEmptyEvt;

fun q_empty (QUEUE (h, t)) = !t=h;
fun q_push (QUEUE (_, t), x) = let val newtail = ref Mnil
			       in
				   !t := Mcons (x, newtail);
				   t := newtail
			       end;
fun q_pop (QUEUE (h, t as ref (Mcons (x, r)))) =
    (h := !r;
     if !t=r
	 then (t := h; x)
     else x)
  | q_pop _ = raise QEmptyEvt;


datatype 'a priority_queue = NOQUEUE
  | BQUEUE of int * 'a * 'a priority_queue * 'a priority_queue; (* this is a heap. *)
*)

(* C1 subsumes C2 iff C1.sigma is included in C2 for some substitution sigma. *)

fun block_subsumes ((pos1, neg1), (pos2, neg2)) = (* b1 subsumes b2 iff b1 is included in b2. *)
    pos1 subset pos2 andalso neg1 subset neg2;

fun eqvar_subsumes (eqvars1, eqvars2) = (* tests whether every equality in eqvars1
					 is realized in eqvars2. *)
    all
      i inset eqvars2 andalso
      js subset ?eqvars2 i
    | i => js in map eqvars1
    end;

(* blocks_subsumes (blocks1, blocks2) detects when
 blocks1 subsumes blocks2, with the empty substitution. *)
fun blocks_subsumes (nil, ...) = true
  | blocks_subsumes (blocks1 as (i,b)::rest, blocks2) =
    case blocks2 of
	nil => false
      | (i',b')::rest' => if i>i'
			      then blocks_subsumes (blocks1, rest')
			  else i=i' andalso
			      block_subsumes (b, b') andalso
			      blocks_subsumes (rest, rest');

fun subsumes (C_BLOCK b1, C_BLOCK b2) = block_subsumes (b1, b2)
  | subsumes (C_BLOCK b1, C_COMPLEX (b2, f, eqvars, blocks)) =
    (* assume block b1 subsumes complex clause C2.  Then either sigma is [x:=f(x1,...,xn)],
     and b1 must subsume (pos2, neg2); or sigma is [x:=y], where x is the variable of b1 and
     y is the variable of one of the one-variable blocks b2 in C2, and b1 subsumes b2. *)
    block_subsumes (b1, b2) orelse
    exists
      block_subsumes (b1, b)
    | (_, b) in list blocks
    end
  | subsumes (C_COMPLEX (b1, f1, eqvars1, blocks1), C_COMPLEX (b2, f2, eqvars2, blocks2)) =
    (* assume C1 subsumes C2.  Then f1=f2, and b1 subsumes b2, and eqvars2 makes all equalities
     in eqvars1 true, and blocks2 is subsumed by fuse_blocks (eqvars2, blocks1). *)
    f1=f2 andalso
    block_subsumes (b1, b2) andalso
    eqvar_subsumes (eqvars1, eqvars2) andalso
    blocks_subsumes (fuse_blocks (eqvars2, blocks1), blocks2)
  | subsumes _ = false; (* a C_COMPLEX clause never subsumes a C_BLOCK clause. *)

fun simple_positives (C_BLOCK (pos, _)) = pos
  | simple_positives (C_COMPLEX (_, _, _, blocks)) =
    union {pos
	  | (_, (pos, _)) in list blocks};

fun simple_negatives (C_BLOCK (_, neg)) = neg
  | simple_negatives (C_COMPLEX (_, _, _, blocks)) =
    union {neg
	  | (_, (_, neg)) in list blocks};

fun clause_subsumed (clauses, pos_index, neg_index, c) =
    exists
      let val (ids, _) = ?pos_index P (* we only need to check blocks with P(x) maximal;
				       indeed all clauses with P(x) maximal must be blocks. *)
      in
	  exists
	    subsumes (?clauses id, c)
	  | id in list ids
	  end
      end
    | P in set pos_index <| simple_positives c
    end orelse
    exists
      let val (ids, _) = ?neg_index P (* we only need to check blocks with P(x) maximal;
				       indeed all clauses with P(x) maximal must be blocks. *)
      in
	  exists
	    subsumes (?clauses id, c)
	  | id in list ids
	  end
      end
    | P in set neg_index <| simple_negatives c
    end orelse
    (case c of
	 C_COMPLEX ((pos, neg), f, eqvars, blocks) =>
	 exists
	   let val (simple, complex) = ?pos_index P
	   in
	       exists
		 subsumes (?clauses id, c)
	       | id in list simple
	       end orelse f inset complex andalso
	       exists
		 subsumes (?clauses id, c)
	       | id in list ?complex f
	       end
	   end
	 | P in set pos_index <| pos
	 end orelse
	 exists
	   let val (simple, complex) = ?neg_index P
	   in
	       exists
		 subsumes (?clauses id, c)
	       | id in list simple
	       end orelse f inset complex andalso
	       exists
		 subsumes (?clauses id, c)
	       | id in list ?complex f
	       end
	   end
	| P in set neg_index <| neg
	 end
       | _ => false);

val clause_cnt = ref 0;
fun gencnt () = (inc clause_cnt; ID_NUM (!clause_cnt));

fun add_clause (cr, prr, pr, nr, dedr, c, ded, P, posq, fopt, present) =
    if clause_subsumed (!cr, !pr, !nr, c)
	then ID_NONE
    else let val id = case ded of
			  J_AX id => if id inset !cr
					 then gencnt ()
				     else id
			| J_DED ded0 => gencnt ()
	     val idxr = if posq then pr else nr
	 in
	     cr := !cr ++ {id => c};
	     (if present then prr := !prr ++ {id => c} else ());
		  (case ded of
		       J_DED ded0 => dedr := !dedr ++ {id => ded0}
		     | _ => ());
		       if P inset !idxr
			   then let val (l, fl) = ?(!idxr) P
				in
				    idxr := !idxr ++ {P =>
						       case fopt of
							   SOME f =>
							   (l, if f inset fl
								   then fl ++ {f => id :: ?fl f}
							       else fl ++ {f => [id]})
							 | _ => (id::l, fl)}
				end
		       else idxr := !idxr ++ {P => case fopt of
						       SOME f => (nil, {f => [id]})
						     | _ => ([id], {})};
			   id
	 end;

fun do_resolve (id, id', c, c', P, sq, cq, posq) = (* where +P(...) occurs in c, -P(...) occurs in c' if posq=true,
						    conversely otherwise. *)
    (let val l = if posq then resolve (c, c', P) else resolve (c', c, P)
	 val ded = if posq then J_DED (id, id', P) else J_DED (id', id, P)
     in
	 case l of
	     [c1] => (case c1 of
			  C_BLOCK ({}, {}) => raise ContradictionEvt
			| _ => sq := (c1, ded) :: !sq)
	   | _ => cq := (l, ded) :: !cq
     end handle NoResolventEvt => ()
	      | TautologyEvt => ())


val tableau_false = nil;
val tableau_true = [BRANCH |[clauses = {}, present = {}, pos_index = {}, neg_index = {},
			      simple_queue = nil, complex_queue = nil, deductions = {}]|];

fun expand_tableau less =
    let val maxlit = maximal_literal less
	fun expand nil = nil
	  | expand (BRANCH |[clauses, present, pos_index, neg_index, simple_queue, complex_queue,
			      deductions]| :: tab) =
	    (let val cr = ref clauses
		 val prr = ref present
		 val pr = ref pos_index
		 val nr = ref neg_index
		 val sq = ref simple_queue
		 val cq = ref complex_queue
		 val dedr = ref deductions
	     in
		 while !sq<>nil do (* only deal with simple queue;
				    we shall deal with complex queue later. *)
		     let val (c, ded)::rest = !sq
			 val ((P, posq), fopt) = maxlit c
			 val id = add_clause (cr, prr, pr, nr, dedr,
					      c, ded, P, posq, fopt, true)
			 val index = if posq then !nr else !pr
		     in
			 sq := rest;
			 if id<>ID_NONE andalso P inset index
			     then let val (simple, complex) = ?index P
				  in
				      iterate
					do_resolve (id, id', c, ?(!prr) id', P, sq, cq, posq)
				      | id' in list simple such that id' inset !prr
				      end;
				      case fopt of
					  SOME f => (* then we must only resolve with clauses having
						     P (f (...)) as maximal atom. *)
					  if f inset complex
					      then
						  iterate
						    do_resolve (id, id', c, ?(!prr) id', P, sq, cq, posq)
						  | id' in list ?complex f such that id' inset !prr
						  end
					  else ()
					| _ => (* here we must resolve with all clauses in complex. *)
					  iterate
					    iterate
					      do_resolve (id, id', c, ?(!prr) id', P, sq, cq, posq)
					    | id' in list fl such that id' inset !prr
					    end
					  | f => fl in map complex
					  end
				  end
			 else ()
		     end;
		 (* Here we have dealt with the simple queue entirely.  It remains to deal
		  with the complex queue, which contains non-trivial disjunctions of blocks. *)
		 case !cq of
		     nil => (* If there is nothing in the complex queue either, then this branch is open,
			     and we terminate. *)
		     BRANCH |[clauses = !cr, present = !prr, pos_index = !pr, neg_index = !nr,
			       simple_queue = nil, complex_queue = !cq, deductions = !dedr]| :: tab
		   | (cl, ded)::cq' =>
		     let val tabr = ref tab
		     in
			 iterate
			   (tabr := BRANCH |[clauses = !cr, present = !prr,
					      pos_index = !pr, neg_index = !nr,
					      simple_queue = [(c, ded)], complex_queue = cq',
					      deductions = !dedr]| :: !tabr;
			    add_clause (cr, prr, pr, nr, dedr, c, ded, P, posq, fopt, false)
			    (* used to avoid redoing work in other branches. *)
			    )
			 | c in set elems cl
			 val ((P, posq), fopt) = maxlit c
			 end;
			 expand (!tabr)
		     end
	     end
		 handle ContradictionEvt => expand tab)
    in
	expand
    end;

fun add_clauses (tab, cl) =
    map (fn (BRANCH (b as |[simple_queue, ...]|)) =>
	    BRANCH (b ++|[simple_queue = simple_queue @ cl]|)) tab;

exception EmptyClause;

fun is_tautology (C_BLOCK (pos, neg)) = pos intersects neg
  | is_tautology (C_COMPLEX ((pos, neg), f, eqvars, blocks)) =
    pos intersects neg orelse
    exists
      pos' intersects neg'
    | (_, (pos', neg')) in list blocks
    end;

fun add_clause_lists (tab, cl) =
    (let val simple = ref nil
	 val complex = ref nil
     in
	 iterate
	   (case l of
		[] => raise EmptyClause
	      | [c] => if is_tautology c
			   then ()
		       else simple := (c, just) :: !simple
	      | _ => if exists is_tautology c | c in list l end
			 then ()
		     else complex := (l, just) :: !complex)
	 | (l, just) in list cl
	 end;
	 map (fn (BRANCH (b as |[simple_queue, complex_queue, ...]|)) =>
		  BRANCH (b ++|[ simple_queue = simple_queue @ !simple,
				 complex_queue = complex_queue @ !complex]|)) tab
     end handle EmptyClause => tableau_false);
