(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

open "lispin_h"
open "trans_h"
open "sort_h"

val decompose_patterns_p = ref false;
val yyin = ref (stdin ++|[close = fn () => ()]|);

fun trans_warn (msg,
		|[first_line, first_column, last_line, last_column, in_file]| : lineinfo) =
    (#put stderr "evatrans: ";
     if in_file<>""
	 then (#put stderr in_file;
	       #put stderr ", line";
	       if first_line<>last_line then #put stderr "s" else ();
	       #put stderr " ";
               print stderr (pack first_line);
	       if first_column<>0 then (#put stderr "("; print stderr (pack first_column);
					#put stderr ")")
		   else ();
	       if first_line<>last_line orelse first_column<>last_column
		   then (#put stderr "-";
			 print stderr (pack last_line);
			 if (last_column<>0) then (#put stderr "("; print stderr (pack last_column);
						   #put stderr ")")
			 else ())
	       else ();
	       #put stderr ": ")
     else ();
     #put stderr msg;
     #put stderr "\n";
     #flush stderr ());

fun trans_fatal msg_linfo = (trans_warn msg_linfo;
			     quit 10);

val trans_err_limit = ref 20;

fun trans_error ((SOME k, msg), linfo) =
    (trans_warn (msg, linfo);
     dec trans_err_limit;
     if !trans_err_limit<=0
	 then quit 10
     else throw k ())
  | trans_error ((_, msg), linfo) =
    trans_fatal (msg, linfo);

fun print_eva_type (f as |[put, ...]|) =
    fn (VAR_TYPE s) => put s
     | (FUN_TYPE |[args, res, hash, secret]|) =>
       let val delim = ref ""
       in
	   put "(";
	   iterate
	     (put (!delim);
	      delim := ",";
	      put l)
	   | l in list args
	   end;
	   put ")";
	   put res;
	   if hash then put " hash" else ();
	   if secret then put " secret" else ()
       end;

val genvarcnt = ref ({} : string -m> int ref);
fun genvar prefix =
    let val nr = if prefix inset !genvarcnt
		     then ?(!genvarcnt) prefix
		 else let val nr = ref 0
		      in
			  genvarcnt := !genvarcnt ++ {prefix => nr};
			  nr
		      end
	val f as |[convert, ...]| = outstring prefix
    in
	inc nr;
	print f (pack (!nr));
	convert ()
    end;
fun genvvar x = genvar (x ^ "-");

val pubk_abbrevs = ref ({} : string -m> string);
val privk_abbrevs = ref ({} : string -m> string);

fun print_eva_term (f as |[put, ...]|) =
    let fun pr (T_VAR x) =
	    put (if x inset !pubk_abbrevs
		     then ?(!pubk_abbrevs) x
		 else if x inset !privk_abbrevs
		     then ?(!privk_abbrevs) x
		 else x)
	  | pr (T_APP ("crypt", [a, m, k])) =
	    (put "{";
	     pr m;
	     put "}_(";
	     pr k;
	     put ")";
	     pr_alg a)
	  | pr (T_APP ("sign", [a, m, k])) =
	    (put "[";
	     pr m;
	     put "]_(";
	     pr k;
	     put ")";
	     pr_alg a)
	  | pr (T_APP ("tuple", l)) = prl l
	  | pr (T_APP ("d", [t])) = pr t
	  | pr (T_APP ("p", [t])) = pr t
	  | pr (T_APP ("a", [t])) = pr t
	  | pr (T_APP ("sa", [t])) = pr t
	  | pr (T_APP ("aa", [t])) = pr t
	  | pr (T_APP ("vanilla", nil)) = put "vanilla"
	  | pr (T_APP ("apply", T_VAR f::l)) = (put f; prl l)
	  | pr (T_APP ("hash-apply", l)) = pr (T_APP ("apply", l))
	  | pr (T_APP ("apply-pubk", alg::T_VAR s::l)) = pr_pubk (s, alg, l)
	  | pr (T_APP ("hash-apply-pubk", alg::T_VAR s::l)) = pr_pubk (s, alg, l)
	  | pr (T_APP ("apply-privk", alg::T_VAR s::l)) = pr_privk (s, alg, l)
	  | pr (T_APP ("hash-apply-privk", alg::T_VAR s::l)) = pr_privk (s, alg, l)
	  | pr (T_APP ("exact", [t])) = (put "<"; pr t; put ">")
	  | pr (T_APP ("located", [T_VAR s, T_VAR p, t])) =
	    (put "("; pr t; put ")"; put "@"; put s; put p)
	  | pr (T_APP (f, l)) = (put f; put "$"; prl l)
	and prl nil = put "()"
	  | prl (a::l) = (put "("; pr a;
			  iterate
			    (put ","; pr t)
			  | t in list l
			  end;
			  put ")")
	and pr_pubk (name, alg, l) =
	    if name inset !pubk_abbrevs
		then (put (?(!pubk_abbrevs) name); pr_alg alg; prl l)
	    else (put "apply-pubk$"; pr_alg alg; prl (T_VAR name::l))
	and pr_privk (name, alg, l) =
	    if name inset !privk_abbrevs
		then (put (?(!privk_abbrevs) name); pr_alg alg; prl l)
	    else (put "apply-privk$"; pr_alg alg; prl (T_VAR name::l))
	and pr_alg (T_APP ("sa", [T_APP ("vanilla", nil)])) = ()
	  | pr_alg a = (put "^("; pr a; put ")")
    in
	pr
    end;

fun illegal_op f = raise FailEvt (NONE, "illegal operator '" ^ f ^ "'");

fun illegal (APP (f, _)) = illegal_op f
  | illegal (VAR x) = raise FailEvt (NONE, "unexpected variable '" ^ x ^ "'")
  | illegal (INT n) = raise FailEvt (NONE,
				     let val f as |[convert, ...]| = outstring "unexpected integer value "
				     in
					 print f (pack n);
					 convert ()
				     end)
  | illegal (LINE (t, _)) = raise FailEvt (NONE, "internal bug [illegal]");

fun warn_secret () =
    (#put stderr "evatrans: the use of 'secret' functions is strongly discouraged, as they may help conceal attacks by assuming the existence of covert channels; use at your own risks.\n");

fun trans_type (LINE (t, linfo)) =
    (trans_type t
     handle FailEvt kmsg => trans_error (kmsg, linfo))
  | trans_type (VAR s) = VAR_TYPE s
  | trans_type (APP ("one-way-function", typ::targs)) =
    (case trans_type typ of
	 VAR_TYPE s => FUN_TYPE |[ args = trans_type_args targs,
				   res = s,
				   hash = true,
				   secret = false
				   ]|
       | _ => raise FailEvt (NONE, "illegal curried function type"))
  | trans_type (APP ("function", typ::targs)) =
    (case trans_type typ of
	 VAR_TYPE s => FUN_TYPE |[ args = trans_type_args targs,
				   res = s,
				   hash = false,
				   secret = false
				   ]|
       | _ => raise FailEvt (NONE, "illegal curried function type"))
  | trans_type (APP ("secret-one-way-function", typ::targs)) =
    (case trans_type typ of
	 VAR_TYPE s => (warn_secret ();
			FUN_TYPE |[ args = trans_type_args targs,
				    res = s,
				    hash = true,
				    secret = true
				    ]|)
       | _ => raise FailEvt (NONE, "illegal curried function type"))
  | trans_type (APP ("secret-function", typ::targs)) =
    (case trans_type typ of
	 VAR_TYPE s => (warn_secret ();
			FUN_TYPE |[ args = trans_type_args targs,
				    res = s,
				    hash = false,
				    secret = true
				    ]|)
       | _ => raise FailEvt (NONE, "illegal curried function type"))
  | trans_type t = illegal t
and trans_type_args targs =
    [(case trans_type typ of
	  VAR_TYPE s => s
	| _ => raise FailEvt (NONE, "illegal higher-order function type"))
    | typ in list targs];

fun free x =
    let fun fr (T_VAR y) = x=y
	  | fr (T_APP (_, l)) = exists fr t | t in list l end
    in
	fr
    end;

fun fvars (T_VAR x) = {x}
  | fvars (T_APP (_, l)) = union {fvars t | t in list l};

fun eva_subst sigma =
    let fun subst (T_VAR x) =
	    if x inset sigma
		then ?sigma x
	    else T_VAR x
	  (*
	  | subst (T_APP ("let", [pat, t])) =
	    T_APP ("let", [patsubst pat, subst t])
	  | subst (T_APP ("new", [pat])) =
	    T_APP ("new", [patsubst pat])
	  | subst (T_APP ("recv", [pat])) =
	    T_APP ("recv", [patsubst pat])
	   *)
	  | subst (T_APP (f, l)) =
	    T_APP (f, [subst t | t in list l])
    (*
	and patsubst (T_APP ("crypt", [a, m, k])) =
	    T_APP ("crypt", [subst a, patsubst m, subst k])
	  | patsubst (T_APP ("exact", [t])) =
	    T_APP ("exact", [subst t])
	  | patsubst (T_APP ("apply", f::l)) =
	    T_APP ("apply", subst f::[patsubst pat | pat in list l])
	  | patsubst (T_APP (f, l)) =
	    T_APP (f, [patsubst pat | pat in list l])
	  | patsubst x = x
     *)
    in
	subst
    end;

fun mcheck (LINE (t, linfo)) =
    (mcheck t handle FailEvt kmsg => trans_error (kmsg, linfo))
  | mcheck (APP ("percent", _)) =
    raise FailEvt (NONE, "illegal nested '%'")
  | mcheck (APP (f, l)) =
    iterate mcheck t | t in list l end
  | mcheck _ = ();

fun msplit (LINE (t, linfo)) =
    (let val (s, r) = msplit t
     in
	 (LINE (s, linfo), LINE (r, linfo))
     end
	 handle FailEvt kmsg => trans_error (kmsg, linfo))
  | msplit (APP ("percent", [s, r])) =
    (mcheck s; mcheck r; (s, r))
  | msplit (APP (f, l)) =
    let val pairs = [msplit t | t in list l]
    in
	(APP (f, [s | (s, _) in list pairs]),
	 APP (f, [t | (_, t) in list pairs]))
    end
  | msplit t = (t, t);

fun complete_coercions coercions =
    let val newcoercions =
	    {t1 => m
	    | t1 => m1 in map coercions
		val m = overwrite [{t3 => f2 o f1
				   | t3 => f2 in map m2}
				  | t2 => f1 in map coercions <| m1
				      val m2 = m1 <-| ?coercions t2]
		such that not (empty m)}
    in
	if empty newcoercions
	    then coercions
	else complete_coercions ((coercions delta newcoercions) ++
				     {t1 => m1 ++ m2
				     | t1 => m1 in map newcoercions <| coercions
					 val m2 = ?newcoercions t1})
    end;

val eva_coercions_ref =
    ref { VAR_TYPE "*D*" => { VAR_TYPE "number" => (fn t => T_APP ("d", [t]))},
	  VAR_TYPE "principal" => { VAR_TYPE "number" => (fn t => T_APP ("p", [t]))},
	  VAR_TYPE "*algo*" => { VAR_TYPE "number" => (fn t => T_APP ("a", [t]))},
	  VAR_TYPE "sym_algo" => { VAR_TYPE "*algo*" => (fn t => T_APP ("sa", [t]))},
	  VAR_TYPE "asym_algo" => { VAR_TYPE "*algo*" => (fn t => T_APP ("aa", [t]))}
	  };

val eva_coercions_completed = ref false

fun eva_coercions () =
    (if !eva_coercions_completed
	 then ()
     else (eva_coercions_ref := complete_coercions (!eva_coercions_ref);
	   eva_coercions_completed := true);
	 !eva_coercions_ref);

fun apply (ltt as [T_VAR _, ...], hash, secret) =
    T_APP (if hash then
	       if secret then "hash-secret-apply"
	       else "hash-apply"
	   else if secret then "secret-apply"
	   else "apply",
	       ltt)
  | apply (T_APP ("lambda-pubk", [alg, f])::lt, hash, _) =
    T_APP (if hash then "hash-apply-pubk" else "apply-pubk", alg::f::lt)
  | apply (T_APP ("lambda-privk", [alg, f])::lt, hash, _) =
    T_APP (if hash then "hash-apply-privk" else "apply-privk", alg::f::lt)
  | apply ([ft, ...], ...) =
    let val f as |[put, convert, ...]| = outstring "cannot apply non-variable '"
    in
	print_eva_term f ft;
	put "'";
	raise FailEvt (NONE, convert ())
    end;

fun locate_error (s, p) = raise FailEvt (NONE, "illegal access to data at " ^ s ^ p);

fun readvar (VAR x) = x
  | readvar (LINE (t, linfo)) =
    (readvar t handle FailEvt kmsg => trans_error (kmsg, linfo))
  | readvar t = illegal t;

fun trans_term_1 (do_locate, env, sigma) =
    let fun do_var check (*;;;*) x =
	    if x inset env
		then let val typ = ?env x
		     in
			 if x inset sigma
			     then (?sigma x, typ)
			 else (check x; (T_VAR x, typ))
		     end
	    else raise FailEvt (NONE, "variable '" ^ x ^ "' undeclared")
	fun is_sym_algo (T_APP ("a", [t])) = is_sym_algo t
	  | is_sym_algo (T_APP ("sa", _)) = true
	  | is_sym_algo _ = false
	fun is_asym_algo (T_APP ("a", [t])) = is_asym_algo t
	  | is_asym_algo (T_APP ("aa", _)) = true
	  | is_asym_algo _ = false
	fun asym_algo (T_APP (f, [id, ...])) =
	    if f inset { "apply-pubk", "hash-apply-pubk",
			 "apply-privk", "hash-apply-privk",
			 "lambda-pubk", "lambda-privk"}
		then SOME id
	    else NONE
	  | asym_algo _ = NONE
	fun algo_name_from_id (T_APP ("a", [t])) = algo_name_from_id t
	  | algo_name_from_id (T_APP ("aa", [t])) = algo_name_from_id t
	  | algo_name_from_id (T_APP ("sa", [t])) = algo_name_from_id t
	  | algo_name_from_id (T_VAR x) = x
	  | algo_name_from_id _ = "<not_found>"
	fun check_sym_asym (m, k, a) =
	    if is_sym_algo a
		then case asym_algo k of
			 SOME id =>
			 let val f as |[put, convert, ...]| =
				 outstring "encryption probably misses algorithm '"
			 in
			     put (algo_name_from_id id);
			     put "'; here symmetric encryption will be assumed, although key is asymmetric";
			     trans_warn (convert (),
					 |[ first_line=0, first_column=0,
					    last_line=0, last_column=0, in_file=""]|)
			 end
		       | NONE => ()
	    else if is_asym_algo a
		then case asym_algo k of
		    SOME id =>
		    if algo_name_from_id a = algo_name_from_id id
			then ()
		    else let val f as |[put, convert, ...]| = outstring "encryption algorithm '"
			 in
			     put (algo_name_from_id a);
			     put "' is different from algorithm '";
			     put (algo_name_from_id id);
			     put "' associated with key";
			     trans_warn (convert (),
					 |[ first_line=0, first_column=0,
					    last_line=0, last_column=0, in_file=""]|)
			 end
		       | NONE =>
			 let val f as |[put, convert, ...]| =
			     outstring "asymmetric encryption, with algorithm '"
			 in
			     put (algo_name_from_id a);
			     put "', used with symmetric key '";
			     put (algo_name_from_id k);
			     put "'";
			     trans_warn (convert (),
					 |[ first_line=0, first_column=0,
					    last_line=0, last_column=0, in_file=""]|)
			 end
	    else ()

	memofun tterm (LINE (t, linfo)) =
		(tterm t handle FailEvt kmsg => trans_error (kmsg, linfo))
	      | tterm (VAR x) =
		(case do_locate of
		     NONE => do_var (fn _ => ()) (*;;;*) x
		   | SOME (_, check_var) => do_var check_var (*;;;*) x)
	      | tterm (APP ("percent", _)) =
		raise FailEvt (NONE, "illegal '%' in message")
	      | tterm (APP ("locate", [vsession, vp, t])) =
		let val session = readvar vsession
		    val p = readvar vp
		in
		    case do_locate of
			SOME (loc, _) =>
			let val (u, typ) = #1 (trans_term_1 (NONE, env, sigma)) t
			in
			    (loc (session, p, u), typ)
			end
		      | _ => locate_error (session, p)
		end
	      | tterm (APP ("crypt", [m, k, a])) =
		let val mt = tterm_t (m, VAR_TYPE "number")
		    val kt = tterm_t (k, VAR_TYPE "number")
		    val at = tterm_t (a, VAR_TYPE "number")
		in
		    check_sym_asym (mt, kt, at);
		    (T_APP ("crypt", [at, mt, kt]),
		     VAR_TYPE "number")
		end
	      | tterm (APP ("sign", [m, k, a])) =
		let val mt = tterm_t (m, VAR_TYPE "number")
		    val kt = tterm_t (k, VAR_TYPE "number")
		    val at = tterm_t (a, VAR_TYPE "number")
		in
		    check_sym_asym (mt, kt, at);
		    (T_APP ("crypt", [at, mt, kt]),
		     VAR_TYPE "number")
		end
	      | tterm (APP ("tuple", l)) =
		let val l' = [tterm_t (t, VAR_TYPE "number") | t in list l]
		in
		    (T_APP ("tuple", l'), VAR_TYPE "number")
		end
	      | tterm (APP ("vanilla", nil)) =
		(T_APP ("vanilla", nil), VAR_TYPE "sym_algo")
	      | tterm (APP (opr as "lambda-pubk", [algo, f])) =
		do_lambda (opr, algo, f)
	      | tterm (APP (opr as "lambda-privk", [algo, f])) =
		do_lambda (opr, algo, f)
	      | tterm (APP ("apply", f::l)) =
		let val (ft, ftyp) = tterm f
		in
		    case ftyp of
			FUN_TYPE |[args, res, hash, secret]| =>
			let val lt = [tterm_t (t, VAR_TYPE typ)
				     || t in list l and
					typ in list args]
				handle ParSweep => raise FailEvt (NONE, "wrong number of arguments")
			in
			    (apply (ft::lt, hash, secret), VAR_TYPE res)
			end
		      | _ => let val f as |[put, convert, ...]| = outstring "cannot apply '"
			     in
				 print_eva_term f ft;
				 put "', of non-functional type '";
				 print_eva_type f ftyp;
				 put "'";
				 raise FailEvt (NONE, convert ())
			     end
		end
	      | tterm t = illegal t
	with tterm_t (LINE (t, linfo), typ) =
	     (tterm_t (t, typ) handle FailEvt kmsg => trans_error (kmsg, linfo))
	   | tterm_t (t, typ) =
	     let val (v, typ') = tterm t
	     in
		 if typ=typ'
		     then v
		 else (?(?(eva_coercions ()) typ') typ v
		       handle MapGet =>
		       let val f as |[put, convert, ...]| = outstring "type mismatch, expected "
			   val pr = print_eva_type f
		       in
			   pr typ;
			   put ", but ";
			   print_eva_term f v;
			   put " is of type ";
			   pr typ';
			   raise FailEvt (NONE, convert ())
		       end)
	     end
	and do_lambda (opr, algo, f) =
	    let val (ft, ftyp) = tterm f
		val alg = tterm_t (algo, VAR_TYPE "*algo*")
	    in
		case ft of
		    T_VAR _ =>
		    (case ftyp of
			 VAR_TYPE s =>
			 raise FailEvt (NONE, "internal bug [tterm], " ^ opr
					^ " has argument of non-functional type '"
					^ s ^ "'")
		       | _ => (T_APP (opr, [alg, ft]), ftyp))
		  | _ => raise FailEvt (NONE, "internal bug [tterm], " ^ opr
					^ " has non-variable argument")
	    end
    in
	(tterm, tterm_t)
    end;

fun trans_term envs = #1 (trans_term_1 envs);
fun trans_term_type envs = #2 (trans_term_1 envs);

(* Note: si on analyse un terme, toutes ses variables libres sont censees etre
 dans type_env; celles qui ne sont pas dans value_env ne doivent jamais pouvoir
 y etre ajoutees: si x n'est pas aliase quand on l'utilise, il ne faut plus
 jamais pouvoir l'aliaser. *)

fun trans_decl (LINE (t, linfo)) =
    (trans_decl t
     handle FailEvt kmsg => trans_error (kmsg, linfo))
  | trans_decl (APP ("declare", l)) =
    let val env = ref ({"**asymk**" => FUN_TYPE |[ args = ["*D*"],
						   res = "number",
						   hash = true,
						   secret = false
						   ]|
			 } : type_env)
	val shared = ref ({} : string set)
	val sigma = ref ({} : value_env)
	val local_sigma = ref ({} : string -m> value_env)
	val everybody = ref ({} : eva_term -m> eva_term)
	val knows = ref ({} : string -m> (eva_term -m> eva_term) ref)
	(* map principal names to sets of terms it knows;
	 each known term is mapped to a private variable standing for the term.
	 Well, this used to be the idea.  Nowadays each known term is simply
	 mapped to itself here.  However other terms may be mapped to different terms,
	 as in function abstract, which gives new names to received data (e.g., if
	 you receive f(x) but don't know f(x), abstract will transform every f(x) term
	 into some new variable).
	 *)
	val axioms = ref ({} : (eva_term * eva_term * eva_type * type_env) set)
	val equations as |[equate, ...]| = cc ()

	val used = ref ({} : string set)

	fun check_var x =
	    if x inset !env
		then raise FailEvt (NONE, "variable '" ^ x ^ "' declared twice")
	    else ()

	fun aliased x =
	    x inset !sigma orelse
	    exists
	      x inset lsigma
	    | _ => lsigma in map !local_sigma
	    end

	fun check_principal x =
	    if x inset !env
		then case ?(!env) x of
			 VAR_TYPE "principal" =>
			 if aliased x
			     then raise FailEvt (NONE, "variable '" ^ x ^ "' is aliased, cannot be used as principal name")
			 else ()
		       | _ => raise FailEvt (NONE, "'" ^ x ^ "' is not a principal name")
	    else raise FailEvt (NONE, "undeclared principal " ^ x)

	fun check_p t = let val x = readvar t
			in
			    check_principal x;
			    x
			end

	fun do_alias (f, x, t) =
	    if aliased x
		then raise FailEvt (NONE, "variable '" ^ x ^ "' already aliased")
	    else if x inset !used
		then raise FailEvt (NONE, "variable '" ^ x
				    ^ "' already used above, cannot be rebound, should have been aliased earlier")
	    else if x inset !env
		then let val tau = ?(!env) x
			 val v = trans_term_type (NONE, !env, !sigma) (t, tau)
		     in
			 if free x v
			     then raise FailEvt (NONE, "variable '" ^ x ^ "' cannot be aliased to term depending on '" ^ x ^ "'")
			 else (f (x, v);
			       used := !used U fvars v)
		     end
	    else let val (v, tau) = trans_term (NONE, !env, !sigma) t
		 in
		     if free x v
			 then raise FailEvt (NONE, "variable '" ^ x ^ "' cannot be aliased to term depending on '" ^ x ^ "'")
		     else (env := !env ++ {x => tau};
			   f (x, v);
			   used := !used U fvars v)
		 end

	fun trans_type_bind (LINE (t, linfo)) =
	    (trans_type_bind t
	     handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trans_type_bind (APP ("type", [vx, typ])) =
	    {readvar vx => trans_type typ}
	  | trans_type_bind t = illegal t

	fun trans_one_decl (LINE (t, linfo)) =
	    (trans_one_decl t
	     handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trans_one_decl (APP ("type", [vx, typ])) =
	    let val x = readvar vx
	    in
		check_var x;
		env := !env ++ {x => trans_type typ}
	    end
	  | trans_one_decl (APP ("type-shared", [vx, typ])) =
	    let val x = readvar vx
	    in
		check_var x;
		env := !env ++ {x => trans_type typ};
		shared := !shared U {x}
	    end
	  | trans_one_decl (APP ("keypair", [algo, vx, vy, typ])) =
	    let val x = readvar vx
		val y = readvar vy
	    in
		check_var x;
		check_var y;
		if x=y
		    then raise FailEvt (NONE, "keypair gives same name (" ^ x ^ ") to both parts of the key")
		else let val alg = trans_term_type (NONE, !env, !sigma) (algo, VAR_TYPE "*algo*")
		     in
			 (case trans_type typ of
			      VAR_TYPE "number" =>
			      let val z = genvar "_kp"
			      in
				  pubk_abbrevs := !pubk_abbrevs ++ {z => x};
				  privk_abbrevs := !privk_abbrevs ++ {z => y};
				  env := !env ++ { z => VAR_TYPE "*D*", (* z has the type *D* of raw data. *)
						   x => VAR_TYPE "number",
						   y => VAR_TYPE "number"};
				  sigma := !sigma ++ { x => T_APP ("hash-apply-pubk", [alg, T_VAR "**asymk**", T_VAR z]),
						       y => T_APP ("hash-apply-privk", [alg, T_VAR "**asymk**", T_VAR z])}
			      end
			    | VAR_TYPE t => raise FailEvt (NONE, "internal bug [trans_one_decl], keypair declares object of non-number type '"
							   ^ t ^ "'")
			    | typ as FUN_TYPE |[args, res = "number", hash = true, secret = false]| =>
			      let val z = genvar "_kpf"
			      in
				  pubk_abbrevs := !pubk_abbrevs ++ {z => x};
				  privk_abbrevs := !privk_abbrevs ++ {z => y};
				  env := !env ++ { z => typ, x => typ, y => typ };
				  shared := !shared U {z};
				  sigma := !sigma ++ { x => T_APP ("lambda-pubk", [alg, T_VAR z]),
						       y => T_APP ("lambda-privk", [alg, T_VAR z])}
			      end
			    | typ as FUN_TYPE |[args, res = "number", hash = true, secret = true]|
			      => raise FailEvt (NONE, "keypair cannot be used with 'secret'")
			    | typ => raise FailEvt (NONE, "keypair on absurd function type"))
		     end
	    end
	  | trans_one_decl (APP ("alias", [vx, t])) =
	    do_alias (fn (x, v) => sigma := !sigma ++ {x => v},
			 readvar vx, t)
	  | trans_one_decl (APP ("local-alias", vx::t::ids)) =
	    do_alias (fn (x, v) =>
			 iterate
			   let val p = check_p id
			   in
			       if p inset !local_sigma
				   then let val lsigma = ?(!local_sigma) p
					in
					    local_sigma := !local_sigma ++ {p => lsigma ++ {x => v}}
					end
			       else local_sigma := !local_sigma ++ {p => {x => v}}
			   end
			 | id in list ids
			 end,
			 readvar vx, t)
	  | trans_one_decl (APP ("basetype", [id])) =
	    (case trans_type id of
		 (typ as VAR_TYPE t) =>
		 if t="number" orelse typ inset !eva_coercions_ref
		     then raise FailEvt (NONE, "cannot declare '" ^ t ^ "' as new basetype")
		 else let val d_coercion = "*" ^ t ^ "_of_D*"
			  val n_coercion = "*number_of_" ^ t ^ "*"
		      in
			  check_var d_coercion;
			  check_var n_coercion;
			  #put stderr "evatrans: the use of 'basetype' (here, 'basetype ";
			  #put stderr t;
			  #put stderr "') is strongly discouraged, as it may help conceal type confusion attacks; use at your own risks.\n";
			  #flush stderr ();
			  env := !env ++ { d_coercion => FUN_TYPE |[ args=["*D*"],
								     res=t,
								     hash=false,
								     secret=false]|,
					   n_coercion => FUN_TYPE |[ args=[t],
								     res="number",
								     hash=false,
								     secret=false]|};
			  everybody := !everybody ++ { T_VAR d_coercion => T_VAR d_coercion,
						       T_VAR n_coercion => T_VAR n_coercion};
			  let val c = !eva_coercions_ref
			      val from_d = ?c (VAR_TYPE "*D*")
			  in
			      eva_coercions_ref := c ++
			      { VAR_TYPE "*D*" => from_d ++ { typ =>
							      (fn t => T_APP ("apply",
									      [ T_VAR d_coercion,
										t])) },
				typ => { VAR_TYPE "number" => (fn t => T_APP ("apply",
									      [ T_VAR n_coercion,
										t])) } };
				eva_coercions_completed := false
			  end
		      end
	       | typ => raise FailEvt (NONE, "cannot declare function type as new base type"))
	  | trans_one_decl (APP ("knows", [vx, t])) =
	    let val x = readvar vx
	    in
		check_principal x;
		let val (v, typ) = trans_term (NONE, !env, !sigma) t
		in
		    if x inset !knows
			then let val k = ?(!knows) x
			     in
				 k := !k ++ {v => v (* y *)}
			     end
		 else knows := !knows ++ {x => ref {v => v (* y *)}};
		     used := !used U {x} U fvars v
		end
	    end
	  | trans_one_decl (APP ("everybody-knows", [t])) =
	    let val (v, typ) = trans_term (NONE, !env, !sigma) t
	    in
		everybody := !everybody ++ { v => v };
		used := !used U fvars v
	    end
	  | trans_one_decl (APP ("knows", [APP ("intruder", nil), t])) =
	    raise FailEvt (NONE, "'intruder knows' clause obsolete")
	  | trans_one_decl (APP ("axiom", s::t::qs)) =
	    let val local_env = overwrite [trans_type_bind t
					  | t in list qs]
		val env' = !env ++ local_env
		val (v, tau) = trans_term (NONE, env', !sigma) t
		val u = trans_term_type (NONE, env', !sigma) (s, tau)
	    in
		axioms := !axioms U {(u, v, tau, local_env)};
		equate (u, v);
		used := !used U fvars u U fvars v
	    end
	  | trans_one_decl t = illegal t
    in
	iterate
	  trans_one_decl decl
	| decl in list l
	end;
	EVA_DECL |[ env = !env,
		    shared = !shared,
		    sigma = !sigma,
		    local_sigma = !local_sigma,
		    everybody = !everybody,
		    knows = {x => !msgs
			    | x => msgs in map !knows},
		    axioms = !axioms,
		    equations = equations
		    ]|
    end
  | trans_decl t = illegal t;

fun trc_do_apply trc =
    let fun trcl l = [trc t | t in list l]
	fun do_apply (T_APP ("hash-apply-pubk", alg::f::l)) =
	    apply (trcl (T_APP ("lambda-pubk", [alg, f])::l), true, false)
	  | do_apply (T_APP ("hash-apply-privk", alg::f::l)) =
	    apply (trcl (T_APP ("lambda-privk", [alg, f])::l), true, true)
	  | do_apply (T_APP ("apply-pubk", alg::f::l)) =
	    apply (trcl (T_APP ("lambda-pubk", [alg, f])::l), false, false)
	  | do_apply (T_APP ("apply-privk", alg::f::l)) =
	    apply (trcl (T_APP ("lambda-privk", [alg, f])::l), false, true)
	  | do_apply (T_APP (f, l)) = T_APP (f, trcl l)
    in
	do_apply
    end;

exception TransComposeEvt of eva_term;

fun trans_compose (env, make, err, |[find, class, ...]|) =
    let val actions = ref nil (* reversed list of actions *)
	val knownr = ref {}
	val unknownr = ref {}
	fun trc t =
	    if t inset !knownr
		then ?(!knownr) t
	    else case t of
		     T_VAR x => 
		     if x inset !unknownr
			 then raise TransComposeEvt t
		     else if x inset env (* compile a 'new' action: *)
			 then case ?env x of
				  (*
				  VAR_TYPE "number" =>
				  let val y = genvvar x
				      val v = T_APP ("d", [T_VAR y])
				  in
				      actions := (T_APP ("new", [T_VAR y]) ::
						  (* allocate y of type *D* *)
						  !actions);
				      knownr := !knownr ++ {t => v (* x *)};
				      v
				  end
				|*) VAR_TYPE "*D*" => (* can only happen when, say,
						     we write Kpub or Kpriv, with
						     keypair Kpub, Kpriv declared
						     earlier; we take this as saying
						     that we must generate the
						     key pair. *)
				  let val y = genvvar x
				      val v = T_VAR y
				  in
				      actions := (T_APP ("new", [v]) ::
						  !actions);
				      knownr := !knownr ++ {t => v};
				      v
				  end
				| (typ as VAR_TYPE _) => (* new on x : typ, where
							  typ is declared of base type;
							  we must have a coercion from
							  *D* to typ. *)
				  (let val c = eva_coercions ()
				       val coerce = ?(?c (VAR_TYPE "*D*")) typ
				       val y = genvvar x
				       val v = coerce (T_VAR y)
				   in
				       actions := (T_APP ("new", [T_VAR y]) ::
						   (* allocate y of type *D* *)
						   !actions);
				       knownr := !knownr ++ {t => v};
				       v
				   end handle MapGet =>
				       let val f as |[put, convert, ...]|
					       = outstring "cannot build new "
				       in
					   put x;
					   put " of type ";
					   print_eva_type f typ;
					   raise FailEvt (NONE, convert ())
				       end)
				| typ => raise TransComposeEvt t
		     else raise FailEvt (NONE, "internal bug [trans_compose]: variable '"
					 ^ x ^ "' undeclared")
		   | _ => trc_do_apply trc t
	fun trc1 (known, unknown, t) = (knownr := known;
					unknownr := unknown;
					actions := nil; trc t)
    in
	fn known_t =>
	   (
(*
#put stderr "trans_compose:\n";
print_eva_term stderr (#2 known_t);
#put stderr "\n";
iterate
  (print_eva_term stderr s;
   #put stderr " = ";
   print_eva_term stderr t;
   #put stderr "\n")
| s => t in map #1 known_t
end;
#put stderr "pubk_abbrevs = ";
pretty stderr (pack !pubk_abbrevs);
#put stderr "privk_abbrevs = ";
pretty stderr (pack !privk_abbrevs);
#put stderr "----------------\n";
#flush stderr ();
*)
	    let val t' = trc1 known_t
		    handle TransComposeEvt m =>
		    let val f as |[put, convert, ...]| =
			    outstring "non-implementable message, sender cannot build '"
		    in
			print_eva_term f m;
			put "'";
			raise err (convert ())
		    end
	    in
		(!knownr, make t' :: !actions)
	    end)
    end;

exception TransExpectTrcEvt of string;

val inverse_key_constr =
    { "hash-apply-pubk" => "hash-apply-privk",
      "hash-apply-privk" => "hash-apply-pubk",
      "apply-pubk" => "apply-privk",
      "apply-privk" => "apply-pubk"
      };

fun unknown_key k =
    let val f as |[put, convert, ...]| = outstring "unimplementable message, key '"
    in
	print_eva_term f k;
	put "' should be asymmetric, is not defined as such to receiver";
	raise FailEvt (NONE, convert ())
    end;

fun bad_algo (a, alg) =
    let val f as |[put, convert, ...]| = outstring "unimplementable message, message is encrypted under algo '"
    in
	print_eva_term f a;
	put "' but key only works for algo '";
	print_eva_term f alg;
	put "'";
	raise FailEvt (NONE, convert ())
    end;

fun make_inv_key (a, k as T_APP (f, alg::l)) =
    if f inset inverse_key_constr
	then let val f' = ?inverse_key_constr f
	     in
		 if a=alg
		     then T_APP (f', alg::l)
		 else bad_algo (a, alg)
	     end
    else unknown_key k
  | make_inv_key (_, k) = unknown_key k;

fun unknown_algo a =
    let val f as |[put, convert, ...]| = outstring "unimplementable message, encryption algorithm '"
    in
	print_eva_term f a;
	put "' is not known to be symmetric or asymmetric to receiver";
	raise FailEvt (NONE, convert ())
    end;

fun trans_comp0 knownr =
    let memofun trc t =
		if t inset !knownr
		    then ?(!knownr) t
		else (case t of
			  T_VAR x => raise TransExpectTrcEvt x
			| _ => trc_do_apply trc t)
    in
	trc
    end;

fun trans_comp1 known =
    let memofun trc t =
		if t inset known
		    then ?known t
		else (case t of
			  T_VAR x => raise TransExpectTrcEvt x
			| _ => trc_do_apply trc t)
    in
	trc
    end;

fun locate sys (s, p, t) =
    let val sname = s ^ p
    in
	if sname inset sys
	    then let val (_, _, known) = ?sys sname
		     val trc = trans_comp1 known
		 in
		     T_APP ("located", [T_VAR s, T_VAR p, trc t])
		     handle TransExpectTrcEvt x =>
		     let val f as |[put, convert, ...]| = outstring ("term '")
		     in
			 print_eva_term f t;
			 put "' cannot be built by ";
			 put sname;
			 put ": '";
			 put x;
			 put "' is not known to ";
			 put p;
			 raise FailEvt (NONE, convert ())
		     end
		 end
	else raise FailEvt (NONE, "unknown session/principal pair " ^ sname)
    end;

fun check_share share x =
    if x inset share
	then ()
    else raise FailEvt (NONE, "variable '" ^ x ^
			"' cannot be used here without saying in which session/role pair");

fun fv (T_VAR x) = {x}
  | fv (T_APP (_, l)) = union {fv t | t in list l};

fun trans_expect (env, |[find, class, ...]|) =
    let val knownr = ref {}
	val trc = trans_comp0 knownr
	fun abstract t =
	    let val x = T_VAR (genvar "_x")
	    in
		
		knownr := !knownr ++ {t => x};
		x
	    end

	fun tre t =
	    (T_APP ("exact", [trc t])
	     handle TransExpectTrcEvt uvar =>
	     (case t of
		  T_VAR x => (* The following simple idea is stupid:
			      (knownr := !knownr ++ {t => t, or x}; t)
			      This is because the "session x=toto" mechanism will then
			      blindly replace variable x by toto, including inside
			      patterns. *)
		  abstract t
		| T_APP ("crypt", [a, m, k]) =>
		  let val a' = trc a
		      val kinv =
			  (case a' of
			       T_APP ("a", [T_APP ("sa", ...)]) =>
			       k (* symmetric encryption algorithm:
				  then we may decrypt using k itself;
				  this works even if k is a key for
				  some other algorithm, even an
				  asymmetric algorithm. *)
			     | T_APP ("a", [alg as T_APP ("aa", ...)]) =>
			       (* asymmetric encryption; must check
				that key k is indeed asymmetric, and
				works with the same algorithm. *)
			       make_inv_key (alg, k)
			     | _ => unknown_algo a')
		  in
		      let val kinv' = trc kinv
		      in
			  T_APP ("crypt", [a', tre m, kinv'])
		      end handle TransExpectTrcEvt _ => abstract t
		  end
		| T_APP ("apply", T_VAR f::l) =>
		  T_APP ("apply", T_VAR f :: [tre t | t in list l])
		| T_APP ("secret-apply", T_VAR f::l) =>
		  T_APP ("secret-apply", T_VAR f :: [tre t | t in list l])
		| T_APP ("apply-pubk", a::T_VAR f::l) => 
		  T_APP ("apply-pubk", trc a::T_VAR f:: [tre t | t in list l])
		| T_APP ("apply-privk", a::T_VAR f::l) => 
		  T_APP ("apply-privk", trc a::T_VAR f:: [tre t | t in list l])
		| T_APP (opr, l) =>
		  if opr inset { "hash-apply", "hash-secret-apply",
				 "hash-apply-pubk", "hash-apply-privk",
				 "apply", "secret-apply", "apply-pubk", "apply-privk", (* should have been dealt above *)
				 "lambda-pubk", "lambda-privk" (* forbidden *)
				 }
		      then abstract t
		  else T_APP (opr, [tre t | t in list l])
		  (* raise FailEvt (NONE, "unimplementable message (unknown how to handle '" ^ opr
		   ^ "' operator)")
		   *)
		  ))

	fun varof (x as T_VAR _) = x
	  | varof t = abstract t

	val actions = ref nil (* reversed list of actions *)
	fun decompose_patterns (x, T_APP ("crypt", [a, m, k])) =
	    let val y = varof m
	    in
		actions := T_APP ("let", [T_APP ("crypt", [a, y, k]), x]) :: !actions;
		decompose_patterns (y, m)
	    end
	  | decompose_patterns (x, T_APP ("apply-pubk", a::f::l)) =
	    let val yl = [varof t | t in list l]
	    in
		actions := T_APP ("let", [T_APP ("apply-pubk", a::f::yl)]) :: !actions;
		iterate
		  decompose_patterns (y, t)
		|| y in list yl and t in list l
		end
	    end
	  | decompose_patterns (x, T_APP ("apply-privk", a::f::l)) =
	    let val yl = [varof t | t in list l]
	    in
		actions := T_APP ("let", [T_APP ("apply-privk", a::f::yl)]) :: !actions;
		iterate
		  decompose_patterns (y, t)
		|| y in list yl and t in list l
		end
	    end
	  | decompose_patterns (x, T_APP ("apply", f::l)) =
	    let val yl = [varof t | t in list l]
	    in
		actions := T_APP ("let", [T_APP ("apply", f::yl)]) :: !actions;
		iterate
		  decompose_patterns (y, t)
		|| y in list yl and t in list l
		end
	    end
	  | decompose_patterns (x, T_APP ("secret-apply", f::l)) =
	    let val yl = [varof t | t in list l]
	    in
		actions := T_APP ("let", [T_APP ("secret-apply", f::yl)]) :: !actions;
		iterate
		  decompose_patterns (y, t)
		|| y in list yl and t in list l
		end
	    end
	  | decompose_patterns (x, t as T_APP ("exact", _)) =
	    actions := T_APP ("let", [t, x]) :: !actions
	  | decompose_patterns (x, T_APP (f, l)) =
	    let val yl = [varof t | t in list l]
	    in
		actions := T_APP ("let", [T_APP (f, yl), x]) :: !actions;
		iterate
		  decompose_patterns (y, t)
		|| y in list yl and t in list l
		end
	    end
	  | decompose_patterns _ = ()
    in
	fn (known, t) => (knownr := known;
			  let val t' = tre t
			  in
			      (!knownr,
			       if !decompose_patterns_p
				   then let val x = varof t'
					in
					    actions := [T_APP ("recv", [x])];
					    decompose_patterns (x, t');
					    !actions
					end
			       else [T_APP ("recv", [t'])])
			  end)
    end;

fun add_action (t : eva_term, prog : eva_program, src : string, tgt : string) : eva_program =
    if src inset prog
	then prog ++ {src => ?prog src U {(tgt, t)}}
    else prog ++ {src => {(tgt, t)}};

val skip_action = T_APP ("skip", nil);

fun add_actions (nil, prog : eva_program, src : string, tgt : string,
		 genlab : unit -> string) : eva_program =
    add_action (skip_action, prog, src, tgt)
  | add_actions (t::rest, prog, src, tgt, genlab) =
    let val lab = genlab ()
    in
	add_actions (rest, add_action (t, prog, lab, tgt),
		     src, lab, genlab)
    end;

fun init_actions (LINE (t, _)) = init_actions t
  | init_actions (APP ("comm", [_, vsender, vreceiver, ...])) =
    {readvar vsender => {}, readvar vreceiver => {}}
  | init_actions (APP ("equals", [vp, ...])) = {readvar vp => {}}
  | init_actions (APP ("assign", [vp, ...])) = {readvar vp => {}}
  | init_actions (APP ("block", l)) = overwrite [init_actions t | t in list l]
  | init_actions (APP ("switch", _ :: l)) = overwrite [init_actions t | t in list l]
  | init_actions (APP ("case", _ :: l)) = overwrite [init_actions t | t in list l]
  | init_actions _ = {};

exception PrincipalNotFound of string;

fun trans_code (env, sigma, local_sigma, everybody,
		knows0, axioms, cc as |[find, add_rewrite, ...]|) =
    let fun tterm_t p = if p inset local_sigma
			    then trans_term_type (NONE, env, ?local_sigma p ++ sigma)
			else trans_term_type (NONE, env, sigma)
	fun tmsg p = let val ttt = tterm_t p
		     in
			 fn t => let val u = ttt (t, VAR_TYPE "number")
				 in
				     find u; (* to add it to the current universe. *)
				     u
				 end
		     end

	fun equals (s, t) =
	    find s=find t orelse
	    (iterate
	       (add_rewrite (u, v, vars)
		handle NotRewrite _ => (add_rewrite (v, u, vars)
					handle NotRewrite _ => () (* ignore equation *)))
	     | (u, v, _, vars) in set axioms
	     end;
	       find s=find t)

	val labels = ref ({} : string set)
	val knows = ref ({} : string -m> eva_knowledge)
	val unknows = ref ({} : string -m> string set)
	val actions = ref ({} : string -m> eva_program)

	val genlabs = ref 0
	fun genlab () =
	    let val f as |[put, convert, ...]| = outstring "%"
	    in
		inc genlabs;
		print f (pack (!genlabs));
		put ".";
		convert ()
	    end

	memofun label (APP ("comm", vlab :: _)) = readvar vlab
	      | label (APP ("block", t :: _)) = label t
	      | label (LINE (t, _)) = label t
	      | label t = genlab ()

	fun equals_or_assign (p, x, msg, f, src, tgt) =
	    if p inset !actions
		then let val known = everybody ++ (if p inset !knows then ?(!knows) p else {})
			 val unknown = if p inset !unknows then ?(!unknows) p else {}
			 val (known', acts) =
			     trans_compose (env, f, fn s => FailEvt (NONE, s), cc)
			     (known, unknown, tmsg p msg)
			 val p_actions = if p inset !actions then ?(!actions) p else raise PrincipalNotFound p
		     in
			 knows := !knows ++ { p => known' };
			 actions := !actions ++ { p => add_actions (acts, p_actions, src, tgt, genlab)}
			 ++ {p' => add_action (skip_action, prog, src, tgt)
			    | p' => prog in map {p} <-| !actions}
		     end
	    else raise PrincipalNotFound p

	val gensw = ref 0
	fun genswitch () = genvar "_sw"

	fun trc (LINE (t, linfo), src : string, tgt : string) =
	    (trc (t, src, tgt)
	     handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trc (APP ("comm", [vlab, vsender, vreceiver, msg]), src, tgt) =
	    let val lab = readvar vlab
		val sender = readvar vsender
		val receiver = readvar vreceiver
	    in
		if lab inset !labels
		    then raise FailEvt (NONE, "duplicate label '" ^ lab ^ "'")
		else if sender=receiver
		    then raise FailEvt (NONE, "cannot send message from principal '" ^ sender ^ "' to itself")
		else let val (s, r) = msplit msg
			 val s_known = everybody ++ (if sender inset !knows
							 then ?(!knows) sender
						     else {})
			 val s_unknown = if sender inset !unknows then ?(!unknows) sender else {}
			 val r_known = everybody ++ (if receiver inset !knows
							 then ?(!knows) receiver
						     else {})
			 val r_unknown = if receiver inset !unknows then ?(!unknows) receiver else {}
			 val smsg = tmsg sender s
			 val rmsg = tmsg receiver r
		     in
			 if equals (smsg, rmsg)
			     then ()
			 else trans_warn ("probable abuse of '%' notation: message " ^ lab ^
					  " is not clearly the same for sender and receiver \
					   \(they \
					   \may be different, or evatrans is not clever enough to \
					   \deduce this from the axioms)",
					  |[first_line=0, first_column=0,
					     last_line=0, last_column=0, in_file=""]|);
			     let val (s_known', s_acts) =
				     trans_compose (env, fn t => T_APP ("send", [t]),
						    fn s => FailEvt (NONE, s), cc)
				     (s_known, s_unknown, smsg)
				 val (r_known', r_acts) =
				     trans_expect (env, cc) (r_known, rmsg)
				 val r_unknown' = r_unknown U fv rmsg
				 val s_actions = if sender inset !actions
						     then ?(!actions) sender
						 else raise PrincipalNotFound sender
				 val r_actions = if receiver inset !actions
						     then ?(!actions) receiver
						 else raise PrincipalNotFound receiver
			     in
				 labels := !labels U {lab};
				 knows := !knows ++ {sender => s_known', receiver => r_known'};
				 unknows := !unknows ++ {receiver => r_unknown'};
				 actions := ({ sender => add_actions (s_acts, s_actions, src, tgt, genlab),
					       receiver => add_actions (r_acts, r_actions, src, tgt, genlab)}
					       ++ {p => add_action (skip_action, prog, src, tgt)
						  | p => prog in map {sender, receiver} <-| !actions})
			     end
		     end
	    end
	  | trc (APP ("block", nil), src, tgt) =
	    actions := !actions ++ {p => add_action (skip_action, prog, src, tgt)
				   | p => prog in map !actions}
	  | trc (APP ("block", [t]), src, tgt) =
	    trc (t, src, tgt)
	  | trc (APP ("block", t :: rest), src, tgt) =
	    let val b = APP ("block", rest)
		val lab = label b
	    in
		trc (t, src, lab);
		trc (b, lab, tgt)
	    end
	  | trc (APP ("equals", [vp, m1, msg]), src, tgt) =
	    let val p = readvar vp
		val x = readvar m1
		val vx = tmsg p m1
	    in
		equals_or_assign (p, x, msg,
				  fn t => T_APP ("let", [T_APP ("exact", [vx]), t]),
				  src, tgt)
	    end
	  | trc (APP ("assign", [cont, vp, vx, msg]), src, tgt) =
	    let val p = readvar vp
		val x = readvar vx
		val y = genvvar x
	    in
		equals_or_assign (p, y, msg,
				  fn t => T_APP ("let", [T_VAR y, t]),
				     src, tgt)
		before let val tx = T_VAR x
			   val known = everybody ++ (if p inset !knows then ?(!knows) p else {})
		       in
			   knows := !knows ++ {p => known ++ {tx => T_VAR y}}
		       end
	    end
	  | trc (APP ("switch", t::cases), src, tgt) = (* liste de (case t cont) *)
	    let exception Switch of string
		val sw = T_VAR (genswitch ())
		val tcomp = trans_compose (env, fn t => T_APP ("let", [sw, t]), Switch, cc)
		val active = {p => known_acts
			     | p in set !actions
			       val known = everybody ++ (if p inset !knows
							     then ?(!knows) p
							 else {})
			       val unknown = if p inset !unknows then ?(!unknows) p else {}
			       val m = tmsg p t
			       val SOME known_acts = SOME (tcomp (known, unknown, m))
				   handle Switch _ => (* message m on which the switch is done
						       cannot be built by principal p; then this
						       switch is compiled as a no-op for p. *)
				   NONE}
		val save_actions = !actions
		val lab1 = genlab ()
		val labcases = [genlab () | _ in list cases]
	    in
		iterate (* compile case switch for each active principal.
			 But do not compile the messages in each branch yet. *)
		  let val p_actions = ?(!actions) p
		      val actsr = ref (add_actions (acts, p_actions, src, lab1, genlab))
		  in
		      knows := !knows ++ { p => known' };
		      iterate
			actsr := add_action (T_APP ("let", [ T_APP ("exact", [trcase_cmp (c, p, known')]),
							     sw]),
					     p_actions, lab1, lab)
		      || c in list cases and lab in list labcases
		      end;
		      actions := !actions ++ { p => !actsr }
		  end
		| p => (known', acts) in map active
		end;
		actions := active <| !actions; (* deal with active principals only, first *)
		iterate
		  trcase (c, lab, tgt)
		|| c in list cases and lab in list labcases
		end handle (x as PrincipalNotFound p) =>
		if p inset save_actions
		    then let val f as |[put, convert, ...]| = outstring "switch branches on '"
			     val m = trans_term_type (NONE, env, sigma) (t, VAR_TYPE "number")
			 in
			     print_eva_term f m;
			     put "', which cannot be built by '";
			     put p;
			     put "', but '";
			     put p;
			     put "' participates in some case of the switch";
			     raise FailEvt (NONE, convert ())
			 end
		else raise x;
		actions := !actions ++ { p => add_action (skip_action, p_actions, src, tgt)
				       | p => p_actions in map active <-| save_actions}
	    end
	  | trc (t, ...) = illegal t
	and trcase_cmp (LINE (t, linfo), p, known) =
	    (trcase_cmp (t, p, known) handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trcase_cmp (APP ("case", [t, ...]), p, known) =
	    let val u = tmsg p t
	    in
		trans_comp1 known u
		handle TransExpectTrcEvt x =>
		let val f as |[put, convert, ...]| = outstring "unimplementable message, '"
		in
		    print_eva_term f u;
		    put "' cannot be built by ";
		    put p;
		    put ": '";
		    put x;
		    put "' is not known to ";
		    put p;
		    raise FailEvt (NONE, convert ())
		end
	    end
	and trcase (LINE (t, linfo), src, tgt) =
	    (trcase (t, src, tgt)
	     handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trcase (APP ("case", _ :: msgs), src, tgt) =
	    trc (APP ("block", msgs), src, tgt)
	  | trcase (t, ...) = illegal t
    in
	fn t => (labels := {"end."}; knows := knows0; actions := init_actions t;
		 let val start = label t
		 in
		     trc (t, start, "end.");
		     (start, {name => (prog,
				       everybody ++ (if name inset !knows
							 then ?(!knows) name
						     else {}))
			     | name => prog in map !actions})
		 end)
    end;

fun benign (T_VAR _) = true
  | benign _ = false;
(* benign (T_APP (f, l)) = f inset {"tuple", "d", "p", "a", "sa", "aa", "vanilla", "apply"}
    andalso all benign t | t in list l end;
  *)

fun trans_sessions (env, sigma) =
    let val tterm_t = trans_term_type (NONE, env, sigma)
	val labels = ref ({} : string -m> string set option * value_env * id_set)

	fun trs_binding (LINE (t, linfo)) =
	    (trs_binding t
	     handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trs_binding (APP ("bind", [vx, t])) =
	    let val x = readvar vx
	    in
		if x inset env
		    then if x inset sigma
			     then raise FailEvt (NONE, "cannot bind aliased variable '" ^ x ^ "'")
			 else let val u = tterm_t (t, ?env x)
			      in
				  if benign u
				      then (x, u)
				  else raise FailEvt (NONE, "cannot bind '" ^ x ^"' to non-variable term (or to variable term of a different type)")
			      end
		else raise FailEvt (NONE, "variable '" ^ x ^"' undeclared")
	    end
	fun trs_bindings (nil, sigma') = sigma'
	  | trs_bindings (bind::rest, sigma') =
	    let val (x, t) = trs_binding bind
	    in
		if x inset sigma'
		    then raise FailEvt (NONE, "cannot bind '" ^ x ^ "' twice")
		else case fvars t <| sigma' of
			 {y => _, ...} => raise FailEvt (NONE, "cannot bind '" ^ x ^
							 "' to term depending on bound variable '" ^ y ^ "'")
		       | _ => trs_bindings (rest, sigma' ++ {x => t})
	    end
	fun check_principals l = {readvar p | p in list l}
	fun trs_principals (LINE (t, linfo)) =
	    (trs_principals t handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trs_principals (APP ("principals-only", l)) =
	    let val ps = check_principals l
	    in
		ID_INCL ps
	    end
	  | trs_principals (APP ("principals-except", l)) =
	    let val ps = check_principals l
	    in
		ID_EXCL ps
	    end
	  | trs_principals t = illegal t
	fun trs_private (LINE (t, linfo)) =
	    (trs_private t handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trs_private (APP ("private", l)) =
	    check_principals l
	fun trs1session (LINE (t, linfo)) =
	    (trs1session t
	     handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trs1session (APP ("session", vlab::principals::l)) =
	    let val lab = readvar vlab
	    in
		if lab inset !labels
		    then raise FailEvt (NONE, "duplicate session label '" ^ lab ^ "'")
		else labels := !labels ++ {lab => (NONE,
						   trs_bindings (l, {}),
						   trs_principals principals)}
	    end
	  | trs1session (APP ("session-repeat", privs::vlab::principals::l)) =
	    let val lab = readvar vlab
	    in
		if lab inset !labels
		    then raise FailEvt (NONE, "duplicate session label '" ^ lab ^ "'")
		else labels := !labels ++ {lab => (SOME (trs_private privs),
						   trs_bindings (l, {}),
						   trs_principals principals)}
	    end
	  | trs1session t = illegal t
	fun trsl nil = nil
	  | trsl (s::rest) =
	    (trs1session s; trsl rest)
	fun trs (LINE (t, linfo)) =
	    (trs t
	     handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trs (APP ("sessions", l)) =
	    trsl l
	  | trs t = illegal t
    in
	fn t => (labels := {}; trs t; !labels)
    end;

fun eva_prog_subst sigma =
    let val tsubst = eva_subst sigma
    in
	fn prog => {src => {(tgt, tsubst t)
			   | (tgt, t) in set m}
		   | src => m in map prog}
    end;

fun known_subst sigma =
    let val tsubst = eva_subst sigma
    in
	fn known => {t => tsubst u
		    | t => u in map known}
    end

fun print_eva_form (f as |[put, ...]|) =
    let val pet = print_eva_term f
	fun prl nil = put "()"
	  | prl (a::l) = (put "("; pet a;
			  iterate
			    (put ","; pet t)
			  | t in list l
			  end;
			  put ")")
	fun pratom (T_VAR x) = put x
	  | pratom (T_APP ("prop", T_VAR p::l)) =
	    (put p; put " "; prl l)
	  | pratom (T_APP ("at", [T_VAR session, T_VAR p, T_VAR lab])) =
	    (put session; put p; put "@"; put lab)
	  | pratom (T_APP ("owns", [T_VAR session, T_VAR p, t])) =
	    (put session; put p; put "("; pet t; put ")")
	  | pratom (T_APP ("secret", [T_VAR session, T_VAR p, t])) =
	    (put "secret("; pet t; put ")")
	  | pratom (T_APP ("not", [F])) = (put "!"; pratom F)
	  | pratom (T_APP ("E", [F])) = (put "*E "; pratom F)
	  | pratom (T_APP ("A", [F])) = (put "*A "; pratom F)
	  | pratom (T_APP ("N", [F])) = (put "*N "; pratom F)
	  | pratom (T_APP ("F", [F])) = (put "*F "; pratom F)
	  | pratom (T_APP ("G", [F])) = (put "*G "; pratom F)
	  | pratom (T_APP ("P", [F])) = (put "*P "; pratom F)
	  | pratom (T_APP ("Q", [F])) = (put "*Q "; pratom F)
	  | pratom (T_APP ("and", nil)) = put "true"
	  | pratom (T_APP ("or", nil)) = put "false"
	  | pratom F = (put "("; primp F; put ")")
	and prand (T_APP ("and", l)) =
	    prand1 l
	  | prand F = pratom F
	and prand1 (F::G::rest) =
	    (pratom F; put " && ";
	     if null rest
		 then prand G
	     else prand1 (G::rest))
	  | prand1 [F] = pratom F
	  | prand1 nil = put "true"
	and pror (T_APP ("or", l)) =
	    pror1 l
	  | pror (T_APP ("U", [F, G])) = (prand F; put " *U "; pror G)
	  | pror (T_APP ("W", [F, G])) = (prand F; put " *W "; pror G)
	  | pror (T_APP ("S", [F, G])) = (prand F; put " *S "; pror G)
	  | pror (T_APP ("B", [F, G])) = (prand F; put " *B "; pror G)
	  | pror F = prand F
	and pror1 (F::G::rest) = 
	    (prand F; put " || ";
	     if null rest
		 then pror G
	     else pror1 (G::rest))
	  | pror1 [F] = prand F
	  | pror1 nil = put "false"
	and primp (T_APP ("implies", [F, G])) =
	    (pror F; put " => "; primp G)
	  | primp F = pror F
    in
	primp
    end;

fun trans_properties (env, shared, sigma, sys) =
    let val tterm_t = trans_term_type (SOME (locate sys, check_share shared), env, sigma)
	(*
	 path-formulae p, state-formulae s:
	 s ::= prop ... | E p | A p | s&&s | s||s | !s | s=>s
	 p ::= s | N p | s U p | s W p | s S p | s B p | F p | G p | P p | Q p | p&&p | p||p | !p | p=>p
	 C'est NCTL* sans les operateurs X (next) et X^{-1}.
	 Au passage, je ne fais plus la difference entre formules d'etat et de chemin
	 syntaxiquement maintenant.
	 *)
	    (*
	fun check_state (F, force_state) =
	    if force_state
		then let val f as |[put, convert, ...]| = outstring "expected state formula, got path formula '"
		     in
			 print_eva_form f F;
			 put "'";
			 raise FailEvt (NONE, convert ())
		     end
	    else F
	     *)
	fun tr_state (LINE (t, linfo)) =
	    (tr_state t handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | tr_state (APP ("prop", vp :: l)) =
	    T_APP ("prop", T_VAR (readvar vp) :: [tterm_t (t, VAR_TYPE "number")
						 | t in list l])
	  | tr_state (APP ("at", [vsession, vp, vlab])) =
	    let val session = readvar vsession
		val p = readvar vp
		val lab = readvar vlab
		val sname = session ^ p
	    in
		if sname inset sys
		    then let val (_, trans, ...) = ?sys sname
			 in
			     if lab inset trans orelse lab="end." orelse
				 exists
				   exists
				     lab=tgt
				   | (tgt, _) in set tgts
				   end
				 | src => tgts in map trans
				 end
				 then T_APP ("at", [T_VAR session, T_VAR p, T_VAR lab])
			     else raise FailEvt (NONE, "unknown label " ^ sname ^ "@" ^ lab)
			 end
		else raise FailEvt (NONE, "unknown session/principal pair " ^ sname)
	    end
	  | tr_state (APP ("owns", [vsession, vp, t])) =
	    let val session = readvar vsession
		val p = readvar vp
		val sname = session ^ p
	    in
		if sname inset sys
		    then T_APP ("owns", [T_VAR session, T_VAR p, tterm_t (t, VAR_TYPE "number")])
		else raise FailEvt (NONE, "unknown session/principal pair " ^ sname)
	    end
	  | tr_state (APP ("secret", [t])) =
	    T_APP ("secret", [tterm_t (t, VAR_TYPE "number")])
	  | tr_state (APP ("implies", [F, G])) = T_APP ("implies", [tr_state F, tr_state G])
	  | tr_state (APP ("and", Fs)) = T_APP ("and", [tr_state F | F in list Fs])
	  | tr_state (APP ("or", Fs)) = T_APP ("or", [tr_state F | F in list Fs])
	  | tr_state (APP ("not", [F])) = T_APP ("not", [tr_state F])
	  | tr_state (APP ("E", [F])) = T_APP ("E", [tr_state F])
	  | tr_state (APP ("A", [F])) = T_APP ("A", [tr_state F])
	  | tr_state (APP ("N", [F])) = T_APP ("N", [tr_state F])
	  | tr_state (APP ("F", [F])) = T_APP ("F", [tr_state F])
	  | tr_state (APP ("G", [F])) = T_APP ("G", [tr_state F])
	  | tr_state (APP ("P", [F])) = T_APP ("P", [tr_state F])
	  | tr_state (APP ("Q", [F])) = T_APP ("Q", [tr_state F])
	  | tr_state (APP ("U", [F,  G])) = T_APP ("U", [tr_state F, tr_state G])
	  | tr_state (APP ("W", [F,  G])) = T_APP ("W", [tr_state F, tr_state G])
	  | tr_state (APP ("S", [F,  G])) = T_APP ("S", [tr_state F, tr_state G])
	  | tr_state (APP ("B", [F,  G])) = T_APP ("B", [tr_state F, tr_state G])
	  | tr_state t = illegal t
    in
	fn l => [tr_state t | t in list l]
    end;

fun elim_skip_action prog =
    let fun reach (src, done) =
	    if src inset prog
		then let val m = ?prog src
			 val skips = {c | c as (_, t) in set m
					    such that t=skip_action}
		     in
			 (m \ skips) U union {reach (tgt, done U {tgt}) | (tgt, _) in set skips}
		     end
	    else {}
    in
	{src => reach (src, {src})
	| src in set prog}
    end;

fun reachable (prog, starts) =
    let val reach = starts U union {{tgt | (tgt, _) in set m}
				   | src => m in map starts <| prog}
    in
	if reach=starts
	    then reach
	else reachable (prog, reach)
    end;

(*
fun coreachable (prog, ends) =
    let val reach = ends U {src
			   | src => m in map prog
			       such that
				   exists
				     tgt inset ends
				   | (tgt, _) in map m
				   end}
    in
	if reach=ends
	    then reach
	else coreachable (prog, reach)
    end;
*)

fun optimize_action (prog, start) =
    let val prog' = elim_skip_action prog
    in
	reachable (prog', {start}) <| prog'
    end;

fun optimize_actions (start, actions) =
    (start, {name => (repeat, optimize_action (prog, start), known)
	    | name => (repeat, prog, known) in map actions});

fun id_set_in (id, ID_INCL s) = id inset s
  | id_set_in (id, ID_EXCL s) = not (id inset s);

fun trans_spec (LINE (t, linfo)) =
    (trans_spec t
     handle FailEvt kmsg => trans_error (kmsg, linfo))
  | trans_spec (APP ("spec", [decl, code, sessions, assumes, claims])) =
    let val EVA_DECL |[ env, shared, sigma, local_sigma,
			everybody, knows, axioms, equations ]|
	    = trans_decl decl
	val (start, actions) = trans_code (env, sigma, local_sigma, everybody,
					   knows, axioms, equations) code
	val sessions = trans_sessions (env, sigma) sessions
	val system = optimize_actions (start,
				       overwrite [{sname ^ rolename => (repeat, ssubst prog, ksubst known)
						  | rolename => (prog, known) in map actions
						    such that id_set_in (rolename, roles)}
						 | sname => (repeat, session, roles) in map sessions
						     val ssubst = eva_prog_subst session
						     val ksubst = known_subst session])
	val trp = trans_properties (env, shared, sigma, #2 system)
	fun tra (LINE (t, linfo)) =
	    (tra t
	     handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | tra (APP ("assume", l)) = trp l
	  | tra t = illegal t
	fun trc (LINE (t, linfo)) =
	    (trc t
	     handle FailEvt kmsg => trans_error (kmsg, linfo))
	  | trc (APP ("claim", l)) = trp l
	  | trc t = illegal t
    in
	EVA_SPEC |[ env=env, sigma=sigma, axioms=axioms, system=system,
		    assume=tra assumes, claim=trc claims ]|
    end
  | trans_spec t = illegal t;

fun translate t =
    (trans_spec t
     handle FailEvt kmsg => trans_error (kmsg, |[ first_line = 0,
						  last_line = 0,
						  first_column = 0,
						  last_column = 0,
						  in_file = "" ]|));

fun convert_eva_term (T_VAR x) = VAR x
  | convert_eva_term (T_APP (f, l)) = APP (f, [convert_eva_term t
					      | t in list l]);

fun convert_eva_type (VAR_TYPE x) = VAR x
  | convert_eva_type (FUN_TYPE |[args, res, hash, secret]|) =
    APP (if hash then if secret then "secret-one-way-function"
		      else "one-way-function"
	 else if secret then "secret-function"
	 else "function",
	     VAR res :: [VAR arg | arg in list args]);

fun convert_type_env1 env =
    [APP ("type", [VAR x, convert_eva_type t])
    | x in list string_sort env
	val t = ?env x];

fun convert_type_env env =
    APP ("types", convert_type_env1 env);

fun convert_value_env sigma =
    APP ("values", [APP ("alias", [VAR x, convert_eva_term t])
		   | x => t in map sigma]);

fun convert_axioms axioms =
    APP ("axioms", [APP ("axiom",
			 convert_eva_term s::
			 convert_eva_term t::
			 convert_eva_type typ::
			 convert_type_env1 env)
		   | (s, t, typ, env) in set axioms]);

fun eva_prog_top_sort (start, prog) =
    let val marked = ref {}
	val list = ref nil
	fun mark node =
	    if node inset !marked
		then ()
	    else (marked := !marked U {node};
		  if node inset prog
		      then
			  (list := node :: !list;
			   iterate
			     mark next
			   | (next, _) in set ?prog node
			   end)
		  else ())
    in
	mark start;
	rev (!list)
    end;

fun convert_known known =
    APP ("knows",
	 [APP ("as", [convert_eva_term t, convert_eva_term u ])
	 | t => u in map known]);

val eva_prog_out_sort = sort (fn ((s1, _), (s2, _)) => s1 strless s2);

fun convert_program (name, start, (repeat, prog, known)) =
    let val l = VAR name :: VAR start ::
	    convert_known known::
	    append [[APP ("trans", [VAR src, VAR tgt, convert_eva_term t])
		    | (tgt, t) in list eva_prog_out_sort m]
		   | src in list eva_prog_top_sort (start, prog)
		       val m = ?prog src]
    in
	case repeat of
	    SOME privs => APP ("repeat-process",
			       APP ("private", [VAR s | s in list string_sort privs])::l)
	  | _ => APP ("process", l)
    end;

fun convert_system (start, sys) =
    APP ("system", [convert_program (name, start, pk)
		   | name in list string_sort sys
		       val pk = ?sys name]);

fun convert_props (kw, l) =
    APP (kw, [convert_eva_term t
	     | t in list l]);

fun convert_eva_spec (EVA_SPEC |[ env, sigma, axioms, system, assume, claim ]|) =
    APP ("compiled-spec", [ convert_type_env env,
			    convert_value_env sigma,
			    convert_axioms axioms,
			    convert_system system,
			    convert_props ("assume", assume),
			    convert_props ("claim", claim)
			    ]);

fun usage () = (#put stderr "Usage: evatrans [file]\n";
		#flush stderr ();
		quit 10);


fun getopt nil = ()
  | getopt ("-d"::l) = (decompose_patterns_p := true; getopt l)
  | getopt (s::l) =
    (let val |[get, getline, close, ...]| = infile s
     in
	 yyin := |[get=get, getline=getline, close=close]|
     end handle IO n => (#put stderr "evatrans: ";
			 #put stderr s;
			 #put stderr ": ";
			 #put stderr (iomsg n);
			 #put stderr "\n";
			 #flush stderr ();
			 quit 10))
  | getopt _ = usage ();

fun main () =
    (getopt (args ());
     let val f as |[close, ...]| = !yyin
	 val t = lispread f before close ()
     in
	 lispwrite stdout (convert_eva_spec (translate t));
	 #put stdout "\n";
	 #flush stdout ()
     end);

